@@ -45,15 +45,40 @@ static CYTHON_INLINE PyObject* __Pyx_Generator_Yield_From(__pyx_CoroutineObject
 
 //////////////////// CoroutineYieldFrom.proto ////////////////////
 
-static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);
+#define __Pyx_Coroutine_Yield_From(gen, source) __Pyx__Coroutine_Yield_From(gen, source, 0)
+static CYTHON_INLINE PyObject* __Pyx__Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source, int warn);
 
 //////////////////// CoroutineYieldFrom ////////////////////
 //@requires: Coroutine
 //@requires: GetAwaitIter
 
-static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
+static int __Pyx_WarnAIterDeprecation(PyObject *aiter) {
+    int result;
+#if PY_MAJOR_VERSION >= 3
+    result = PyErr_WarnFormat(
+        PyExc_PendingDeprecationWarning, 1,
+        "'%.100s' implements legacy __aiter__ protocol; "
+        "__aiter__ should return an asynchronous "
+        "iterator, not awaitable",
+        Py_TYPE(aiter)->tp_name);
+#else
+    result = PyErr_WarnEx(
+        PyExc_PendingDeprecationWarning,
+        "object implements legacy __aiter__ protocol; "
+        "__aiter__ should return an asynchronous "
+        "iterator, not awaitable",
+        1);
+#endif
+    return result != 0;
+}
+
+static CYTHON_INLINE PyObject* __Pyx__Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source, int warn) {
     PyObject *retval;
     if (__Pyx_Coroutine_CheckExact(source)) {
+        if (warn && unlikely(__Pyx_WarnAIterDeprecation(source))) {
+            /* Warning was converted to an error. */
+            return NULL;
+        }
         retval = __Pyx_Generator_Next(source);
         if (retval) {
             Py_INCREF(source);
@@ -64,6 +89,11 @@ static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject
         PyObject *source_gen = __Pyx__Coroutine_GetAwaitableIter(source);
         if (unlikely(!source_gen))
             return NULL;
+        if (warn && unlikely(__Pyx_WarnAIterDeprecation(source))) {
+            /* Warning was converted to an error. */
+            Py_DECREF(source_gen);
+            return NULL;
+        }
         // source_gen is now the iterator, make the first next() call
         if (__Pyx_Coroutine_CheckExact(source_gen)) {
             retval = __Pyx_Generator_Next(source_gen);
@@ -80,6 +110,53 @@ static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject
 }
 
 
+//////////////////// CoroutineAIterYieldFrom.proto ////////////////////
+
+static CYTHON_INLINE PyObject* __Pyx_Coroutine_AIter_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);
+
+//////////////////// CoroutineAIterYieldFrom ////////////////////
+//@requires: CoroutineYieldFrom
+
+static CYTHON_INLINE PyObject* __Pyx_Coroutine_AIter_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
+#if PY_MAJOR_VERSION >= 3
+    __Pyx_PyAsyncMethodsStruct* am = __Pyx_PyType_AsAsync(source);
+    if (likely(am && am->am_anext)) {
+        // Starting with CPython 3.5.2, __aiter__ should return
+        // asynchronous iterators directly (not awaitables that
+        // resolve to asynchronous iterators.)
+        //
+        // Therefore, we check if the object that was returned
+        // from __aiter__ has an __anext__ method.  If it does,
+        // we return it directly as StopIteration result,
+        // which avoids yielding.
+        //
+        // See http://bugs.python.org/issue27243 for more
+        // details.
+        PyErr_SetObject(PyExc_StopIteration, source);
+        return NULL;
+    }
+#endif
+#if PY_VERSION_HEX < 0x030500B2
+    if (!__Pyx_PyType_AsAsync(source)) {
+        #ifdef __Pyx_Coroutine_USED
+        if (!__Pyx_Coroutine_CheckExact(source))  // quickly rule out a likely case
+        #endif
+        {
+            // same as above in slow
+            PyObject *method = __Pyx_PyObject_GetAttrStr(source, PYIDENT("__anext__"));
+            if (method) {
+                Py_DECREF(method);
+                PyErr_SetObject(PyExc_StopIteration, source);
+                return NULL;
+            }
+            PyErr_Clear();
+        }
+    }
+#endif
+    return __Pyx__Coroutine_Yield_From(gen, source, 1);
+}
+
+
 //////////////////// GetAwaitIter.proto ////////////////////
 
 static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o); /*proto*/
@@ -196,7 +273,7 @@ static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAsyncIter(PyObject *obj) {
             return NULL;
     }
 #else
-    // avoid 'unused function' warning
+    // avoid C warning about 'unused function'
     if (0) (void) __Pyx_PyObject_CallMethod0(obj, PYIDENT("__aiter__"));
 #endif
 
@@ -238,6 +315,8 @@ static void __Pyx_Generator_Replace_StopIteration(void) {
     PyObject *exc, *val, *tb;
     // Chain exceptions by moving StopIteration to exc_info before creating the RuntimeError.
     // In Py2.x, no chaining happens, but the exception still stays visible in exc_info.
+    __Pyx_PyThreadState_declare
+    __Pyx_PyThreadState_assign
     __Pyx_GetException(&exc, &val, &tb);
     Py_XDECREF(exc);
     Py_XDECREF(val);
@@ -262,13 +341,15 @@ typedef struct {
     PyObject *yieldfrom;
     PyObject *gi_name;
     PyObject *gi_qualname;
+    PyObject *gi_modulename;
     int resume_label;
     // using T_BOOL for property below requires char value
     char is_running;
 } __pyx_CoroutineObject;
 
-static __pyx_CoroutineObject *__Pyx__Coroutine_New(PyTypeObject *type, __pyx_coroutine_body_t body,
-                                                   PyObject *closure, PyObject *name, PyObject *qualname); /*proto*/
+static __pyx_CoroutineObject *__Pyx__Coroutine_New(
+    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *closure,
+    PyObject *name, PyObject *qualname, PyObject *module_name); /*proto*/
 static int __Pyx_Coroutine_clear(PyObject *self); /*proto*/
 
 #if 1 || PY_VERSION_HEX < 0x030300B0
@@ -285,8 +366,8 @@ static PyTypeObject *__pyx_CoroutineType = 0;
 static PyTypeObject *__pyx_CoroutineAwaitType = 0;
 #define __Pyx_Coroutine_CheckExact(obj) (Py_TYPE(obj) == __pyx_CoroutineType)
 
-#define __Pyx_Coroutine_New(body, closure, name, qualname)  \
-    __Pyx__Coroutine_New(__pyx_CoroutineType, body, closure, name, qualname)
+#define __Pyx_Coroutine_New(body, closure, name, qualname, module_name)  \
+    __Pyx__Coroutine_New(__pyx_CoroutineType, body, closure, name, qualname, module_name)
 
 static int __pyx_Coroutine_init(void); /*proto*/
 static PyObject *__Pyx__Coroutine_await(PyObject *coroutine); /*proto*/
@@ -298,15 +379,17 @@ static PyObject *__Pyx__Coroutine_await(PyObject *coroutine); /*proto*/
 static PyTypeObject *__pyx_GeneratorType = 0;
 #define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
 
-#define __Pyx_Generator_New(body, closure, name, qualname)  \
-    __Pyx__Coroutine_New(__pyx_GeneratorType, body, closure, name, qualname)
+#define __Pyx_Generator_New(body, closure, name, qualname, module_name)  \
+    __Pyx__Coroutine_New(__pyx_GeneratorType, body, closure, name, qualname, module_name)
 
 static PyObject *__Pyx_Generator_Next(PyObject *self);
 static int __pyx_Generator_init(void); /*proto*/
 
 
 //////////////////// CoroutineBase ////////////////////
+//@substitute: naming
 //@requires: Exceptions.c::PyErrFetchRestore
+//@requires: Exceptions.c::PyThreadStateGet
 //@requires: Exceptions.c::SwapException
 //@requires: Exceptions.c::RaiseException
 //@requires: ObjectHandling.c::PyObjectCallMethod1
@@ -332,6 +415,8 @@ static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
 static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue) {
     PyObject *et, *ev, *tb;
     PyObject *value = NULL;
+    __Pyx_PyThreadState_declare
+    __Pyx_PyThreadState_assign
 
     __Pyx_ErrFetch(&et, &ev, &tb);
 
@@ -450,6 +535,7 @@ int __Pyx_Coroutine_CheckRunning(__pyx_CoroutineObject *gen) {
 static CYTHON_INLINE
 PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value) {
     PyObject *retval;
+    __Pyx_PyThreadState_declare
 
     assert(!self->is_running);
 
@@ -467,7 +553,7 @@ PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value) {
         return NULL;
     }
 
-
+    __Pyx_PyThreadState_assign
     if (value) {
 #if CYTHON_COMPILING_IN_PYPY
         // FIXME: what to do in PyPy?
@@ -475,13 +561,12 @@ PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value) {
         // Generators always return to their most recent caller, not
         // necessarily their creator.
         if (self->exc_traceback) {
-            PyThreadState *tstate = PyThreadState_GET();
             PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
             PyFrameObject *f = tb->tb_frame;
 
-            Py_XINCREF(tstate->frame);
+            Py_XINCREF($local_tstate_cname->frame);
             assert(f->f_back == NULL);
-            f->f_back = tstate->frame;
+            f->f_back = $local_tstate_cname->frame;
         }
 #endif
         __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
@@ -560,8 +645,7 @@ static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
         #endif
         {
             if (value == Py_None)
-                // FIXME - is this the right thing to do?
-                ret = PyIter_Next(yf);
+                ret = Py_TYPE(yf)->tp_iternext(yf);
             else
                 ret = __Pyx_PyObject_CallMethod1(yf, PYIDENT("send"), value);
         }
@@ -799,6 +883,7 @@ static void __Pyx_Coroutine_del(PyObject *self) {
     PyObject *res;
     PyObject *error_type, *error_value, *error_traceback;
     __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
+    __Pyx_PyThreadState_declare
 
     if (gen->resume_label <= 0)
         return ;
@@ -810,6 +895,7 @@ static void __Pyx_Coroutine_del(PyObject *self) {
 #endif
 
     // Save the current exception, if any.
+    __Pyx_PyThreadState_assign
     __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
 
     res = __Pyx_Coroutine_Close(self);
@@ -914,8 +1000,9 @@ __Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value)
     return 0;
 }
 
-static __pyx_CoroutineObject *__Pyx__Coroutine_New(PyTypeObject* type, __pyx_coroutine_body_t body,
-                                                   PyObject *closure, PyObject *name, PyObject *qualname) {
+static __pyx_CoroutineObject *__Pyx__Coroutine_New(
+            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *closure,
+            PyObject *name, PyObject *qualname, PyObject *module_name) {
     __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
 
     if (gen == NULL)
@@ -936,6 +1023,8 @@ static __pyx_CoroutineObject *__Pyx__Coroutine_New(PyTypeObject* type, __pyx_cor
     gen->gi_qualname = qualname;
     Py_XINCREF(name);
     gen->gi_name = name;
+    Py_XINCREF(module_name);
+    gen->gi_modulename = module_name;
 
     PyObject_GC_Track(gen);
     return gen;
@@ -952,11 +1041,7 @@ typedef struct {
 } __pyx_CoroutineAwaitObject;
 
 static void __Pyx_CoroutineAwait_dealloc(PyObject *self) {
-#if CYTHON_COMPILING_IN_CPYTHON
-    _PyObject_GC_UNTRACK(self);
-#else
     PyObject_GC_UnTrack(self);
-#endif
     Py_CLEAR(((__pyx_CoroutineAwaitObject*)self)->coroutine);
     PyObject_GC_Del(self);
 }
@@ -1077,7 +1162,7 @@ static CYTHON_INLINE PyObject *__Pyx__Coroutine_await(PyObject *coroutine) {
     Py_INCREF(coroutine);
     await->coroutine = coroutine;
 #if CYTHON_COMPILING_IN_CPYTHON
-    _PyObject_GC_TRACK(await);
+    PyObject_GC_Track(await);
 #endif
     return (PyObject*)await;
 }
@@ -1099,7 +1184,7 @@ static void __Pyx_Coroutine_check_and_dealloc(PyObject *self) {
 #if PY_VERSION_HEX >= 0x03030000 || defined(PyErr_WarnFormat)
         PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname);
 #else
-        PyObject *msg;
+        {PyObject *msg;
         char *cmsg;
         #if CYTHON_COMPILING_IN_PYPY
         msg = NULL;
@@ -1140,7 +1225,7 @@ static void __Pyx_Coroutine_check_and_dealloc(PyObject *self) {
         #endif
         if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
             PyErr_WriteUnraisable(self);
-        Py_XDECREF(msg);
+        Py_XDECREF(msg);}
 #endif
         PyObject_GC_Track(self);
     }
@@ -1180,6 +1265,7 @@ static PyMemberDef __pyx_Coroutine_memberlist[] = {
     {(char *) "cr_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
     {(char*) "cr_await", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
      (char*) PyDoc_STR("object being awaited, or None")},
+    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), PY_WRITE_RESTRICTED, 0},
     {0, 0, 0, 0, 0}
 };
 
@@ -1387,6 +1473,7 @@ static void __Pyx__ReturnWithStopIteration(PyObject* value); /*proto*/
 
 /////////////// ReturnWithStopIteration ///////////////
 //@requires: Exceptions.c::PyErrFetchRestore
+//@requires: Exceptions.c::PyThreadStateGet
 //@substitute: naming
 
 // 1) Instantiating an exception just to pass back a value is costly.
@@ -1397,6 +1484,7 @@ static void __Pyx__ReturnWithStopIteration(PyObject* value); /*proto*/
 static void __Pyx__ReturnWithStopIteration(PyObject* value) {
     PyObject *exc, *args;
 #if CYTHON_COMPILING_IN_CPYTHON
+    __Pyx_PyThreadState_declare
     if ((PY_VERSION_HEX >= 0x03030000 && PY_VERSION_HEX < 0x030500B1) || PyTuple_Check(value)) {
         args = PyTuple_New(1);
         if (unlikely(!args)) return;
@@ -1410,7 +1498,8 @@ static void __Pyx__ReturnWithStopIteration(PyObject* value) {
         Py_INCREF(value);
         exc = value;
     }
-    if (!PyThreadState_GET()->exc_type) {
+    __Pyx_PyThreadState_assign
+    if (!$local_tstate_cname->exc_type) {
         // no chaining needed => avoid the overhead in PyErr_SetObject()
         Py_INCREF(PyExc_StopIteration);
         __Pyx_ErrRestore(PyExc_StopIteration, exc, NULL);
@@ -1701,11 +1790,7 @@ static PyTypeObject __Pyx__PyExc_StopAsyncIteration_type = {
     0,                                  /*tp_print*/
     0,                                  /*tp_getattr*/
     0,                                  /*tp_setattr*/
-#if PY_MAJOR_VERSION < 3
-    0,                                  /*tp_compare*/
-#else
-    0,                                  /*reserved*/
-#endif
+    0,                                  /*tp_compare / reserved*/
     0,                                  /*tp_repr*/
     0,                                  /*tp_as_number*/
     0,                                  /*tp_as_sequence*/
